/*
 * Copyright (C) 2012 Google Inc.
 * Licensed to The Android Open Source Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 
var ENABLE_MUNGE_TABLES = true;
var ENABLE_MUNGE_IMAGES = true;
var ENABLE_CONTENT_READY = true;
var NORMALIZE_MESSAGE_WIDTHS = true;

var BLOCKED_SRC_ATTR = "blocked-src";

// the set of Elements currently scheduled for processing in handleAllImageLoads
// this is an Array, but we treat it like a Set and only insert unique items
var gImageLoadElements = [];

var gScaleInfo;

/**
 * Only revert transforms that do an imperfect job of shrinking content if they fail
 * to shrink by this much. Expressed as a ratio of:
 * (original width difference : width difference after transforms);
 */
var TRANSFORM_MINIMUM_EFFECTIVE_RATIO = 0.7;
var ZOOM_MINIMUM_EFFECTIVE_RATIO = 0.1;

// Don't ship with this on.
var DEBUG_DISPLAY_TRANSFORMS = true;

var gTransformText = {};

var shouldRunOnResize = true;
var oldScrollWidth = -1;

function isConversationEmpty(bodyDivs) {
    var i, len;
    var msgBody;
    var text;

    // Check if given divs are empty (in appearance), and disable zoom if so.
    for (i = 0, len = bodyDivs.length; i < len; i++) {
        msgBody = bodyDivs[i];
        // use 'textContent' to exclude markup when determining whether bodies are empty
        // (fall back to more expensive 'innerText' if 'textContent' isn't implemented)
        text = msgBody.textContent || msgBody.innerText;
        if (text.trim().length > 0) {
            console.log("The body is empty");
            return false;
        }
    }
    return true;
}

function normalizeAllMessageWidths() {
    var expandedBodyDivs;
    var metaViewport;
    var contentValues;
    var isEmpty;

    expandedBodyDivs = document.querySelectorAll(".mail-message-content");

    console.log("normalizeAllMessageWidths(): body elements = " + expandedBodyDivs.length);

    normalizeElementWidths(expandedBodyDivs);
}

/*
 * Normalizes the width of all elements supplied to the document body's overall width.
 * Narrower elements are zoomed in, and wider elements are zoomed out.
 * This method is idempotent.
 */
function normalizeElementWidths(elements) {
    var i;
    var el;
    var documentWidth;
    var newZoom, oldZoom;

    //documentWidth = document.body.offsetWidth;
	
	documentWidth = WEBVIEW_WIDTH;

    for (i = 0; i < elements.length; i++) {
        el = elements[i];
        oldZoom = el.style.zoom;
        // reset any existing normalization
        if (oldZoom) {
            el.style.zoom = 1;
        }
		
        newZoom = documentWidth / el.scrollWidth;
        transformContent(el, documentWidth, el.scrollWidth);
        newZoom = documentWidth / el.scrollWidth;
		console.log("oldZoom = " + oldZoom + " newZoom = " + newZoom );
		
		el.style.zoom = newZoom;        
    }

    console.log("normalizeElementWidths() completed div w = " + el.scrollWidth + "body w = " + document.body.scrollWidth);    

    // Save the scroll width
    oldScrollWidth = document.body.scrollWidth;
}

function transformContent(el, docWidth, elWidth) {
    var nodes;
    var i, len;
    var index;
    var newWidth = elWidth;
    var wStr;
    var touched;
    // the format of entries in this array is:
    // entry := [ undoFunction, undoFunctionThis, undoFunctionParamArray ]
    var actionLog = [];
    var node;
    var done = false;
    var msgId;
    var transformText;
    var existingText;
    var textElement;
    var start;
    var beforeWidth;
    var tmpActionLog = [];
    if (elWidth <= docWidth) {
        console.log("Already normalized: elWidth = " + elWidth + " docWidth = " + docWidth);
        return;
    }

    console.log("started transform - screen: w=" + WEBVIEW_WIDTH + " scroll: w=" + elWidth + "body w=" + document.body.scrollWidth);

    start = Date.now();

    if (el.parentElement.classList.contains("mail-message")) {
        msgId = el.parentElement.id;
        transformText = "[origW=" + elWidth + "/" + docWidth;
    }

    // Try munging all divs or textareas with inline styles where the width
    // is wider than docWidth, and change it to be a max-width.
    touched = false;
    nodes = ENABLE_MUNGE_TABLES ? el.querySelectorAll("div[style], textarea[style]") : [];
	console.log("before div transform nodes: " + nodes.length);
    touched = transformBlockElements(nodes, docWidth, actionLog);
    if (touched) {
        newWidth = el.scrollWidth;
        console.log("ran div-width munger on el=" + el + " oldW=" + elWidth + " newW=" + newWidth
            + " docW=" + docWidth);
        if (msgId) {
            transformText += " DIV:newW=" + newWidth;
        }
        if (newWidth <= docWidth) {
            done = true;
        }
    }

    if (!done) {
        // OK, that wasn't enough. Find images with widths and override their widths.
        nodes = ENABLE_MUNGE_IMAGES ? el.querySelectorAll("img") : [];
		console.log("before img transform nodes: " + nodes.length);
        touched = transformImages(nodes, docWidth, actionLog);
        if (touched) {
            newWidth = el.scrollWidth;
            console.log("ran img munger on el=" + el + " oldW=" + elWidth + " newW=" + newWidth
                + " docW=" + docWidth);
            if (msgId) {
                transformText += " IMG:newW=" + newWidth;
            }
            if (newWidth <= docWidth) {
                done = true;
            }
        }
    }

    if (!done) {
        // OK, that wasn't enough. Find tables with widths and override their widths.
        // Also ensure that any use of 'table-layout: fixed' is negated, since using
        // that with 'width: auto' causes erratic table width.
        nodes = ENABLE_MUNGE_TABLES ? el.querySelectorAll("table") : [];
        touched = addClassToElements(nodes, shouldMungeTable, "munged",
            actionLog);
        if (touched) {
            newWidth = el.scrollWidth;
            console.log("ran table munger on el=" + el + " oldW=" + elWidth + " newW=" + newWidth
                + " docW=" + docWidth);
            if (msgId) {
                transformText += " TABLE:newW=" + newWidth;
            }
            if (newWidth <= docWidth) {
                done = true;
            }
        }
    }

    if (!done) {
        // OK, that wasn't enough. Try munging all <td> to override any width and nowrap set.
        beforeWidth = newWidth;
        nodes = ENABLE_MUNGE_TABLES ? el.querySelectorAll("td") : [];
        touched = addClassToElements(nodes, null /* mungeAll */, "munged",
            tmpActionLog);
        if (touched) {
            newWidth = el.scrollWidth;
            console.log("ran td munger on el=" + el + " oldW=" + elWidth + " newW=" + newWidth
                + " docW=" + docWidth);
            if (msgId) {
                transformText += " TD:newW=" + newWidth;
            }
            if (newWidth <= docWidth) {
                done = true;
            } else if (newWidth == beforeWidth) {
                // this transform did not improve things, and it is somewhat risky.
                // back it out, since it's the last transform and we gained nothing.
				console.log("before undo risky munging nodes: " + tmpActionLog.length);
                undoActions(tmpActionLog);
            } else {
                // the transform WAS effective (although not 100%)
                // copy the temporary action log entries over as normal
                for (i = 0, len = tmpActionLog.length; i < len; i++) {
                    actionLog.push(tmpActionLog[i]);
                }
            }
        }
    }

    // If the transformations shrank the width significantly enough, leave them in place.
    // We figure that in those cases, the benefits outweight the risk of rendering artifacts.
    if (!done && (elWidth - newWidth) / (elWidth - docWidth) >
            TRANSFORM_MINIMUM_EFFECTIVE_RATIO) {
        console.log("transform(s) deemed effective enough");
        done = true;
    }

    if (done) {
        if (msgId) {
            transformText += "]";
            existingText = gTransformText[msgId];
            if (!existingText) {
                transformText = "Message transforms: " + transformText;
            } else {
                transformText = existingText + " " + transformText;
            }
            gTransformText[msgId] = transformText;
            
            if (DEBUG_DISPLAY_TRANSFORMS) {
                console.log(transformText);
            }
        }
        console.log("munger(s) succeeded, elapsed time=" + (Date.now() - start));
        return;
    }

    // reverse all changes if the width is STILL not narrow enough
    // (except the width->maxWidth change, which is not particularly destructive)
    undoActions(actionLog);
    if (actionLog.length > 0) {
        console.log("all mungers failed, changes reversed. elapsed time=" + (Date.now() - start));
    }
}

function undoActions(actionLog) {
    for (i = 0, len = actionLog.length; i < len; i++) {
        actionLog[i][0].apply(actionLog[i][1], actionLog[i][2]);
    }
}

function addClassToElements(nodes, conditionFn, classToAdd, actionLog) {
    var i, len;
    var node;
    var added = false;
    for (i = 0, len = nodes.length; i < len; i++) {
        node = nodes[i];
        if (!conditionFn || conditionFn(node)) {
            if (node.classList.contains(classToAdd)) {
                continue;
            }
            node.classList.add(classToAdd);
            added = true;
            actionLog.push([node.classList.remove, node.classList, [classToAdd]]);
        }
    }
    return added;
}

function transformBlockElements(nodes, docWidth, actionLog) {
    var i, len;
    var node;
    var wStr;
    var index;
    var touched = false;

    for (i = 0, len = nodes.length; i < len; i++) {
        node = nodes[i];
        wStr = node.style.width || node.style.minWidth;
        index = wStr ? wStr.indexOf("px") : -1;
        if (index >= 0 && wStr.slice(0, index) > docWidth) {
            saveStyleProperty(node, "width", actionLog);
            saveStyleProperty(node, "minWidth", actionLog);
            saveStyleProperty(node, "maxWidth", actionLog);
            node.style.width = "100%";
            node.style.minWidth = "";
            node.style.maxWidth = wStr;
            touched = true;
        }
    }
    return touched;
}

function transformImages(nodes, docWidth, actionLog) {
    var i, len;
    var node;
    var w, h;
    var touched = false;

    for (i = 0, len = nodes.length; i < len; i++) {
        node = nodes[i];
        w = node.offsetWidth;
        h = node.offsetHeight;
        // shrink w/h proportionally if the img is wider than available width
		//node.removeAttribute("title");
		node.removeAttribute("alt");
        if (w > docWidth) {
            saveStyleProperty(node, "maxWidth", actionLog);
            saveStyleProperty(node, "width", actionLog);
            saveStyleProperty(node, "height", actionLog);
            node.style.maxWidth = docWidth + "px";
            node.style.width = "100%";
            node.style.height = "auto";			
            touched = true;
        }
    }
    return touched;
}

function saveStyleProperty(node, property, actionLog) {
    var savedName = "data-" + property;
    node.setAttribute(savedName, node.style[property]);
    actionLog.push([undoSetProperty, node, [property, savedName]]);
}

function undoSetProperty(property, savedProperty) {
    this.style[property] = savedProperty ? this.getAttribute(savedProperty) : "";
}

function shouldMungeTable(table) {
    return table.hasAttribute("width") || table.style.width;
}

// Called when the page is loaded
function onPageLoaded() {
    console.log("onload w=" + window.innerWidth + " h=" + window.innerHeight);
    console.log("body w=" + document.body.scrollWidth + " oldScrollWidth=" + oldScrollWidth);

    // When the page is loaded we need to call the normilize again. This usualy happens after all of the images were loaded.
    // We will check if the width of the scroll was changed in more than 10% (In case the script was running on begging)
    if (window.innerWidth != 0 && document.body.scrollWidth > oldScrollWidth * 1.1) {
		if(!SHOW_PICTURES) {
			// In case we are not in show pictures mode we shouldn't run the script again on resize event
			shouldRunOnResize = false;
		}
        normalizeAllMessageWidths();		
    }
}

// Called on browser resize
function onPageResize() {
    console.log("onresize w=" + window.innerWidth + " h=" + window.innerHeight);

    if (!shouldRunOnResize) {
        // No need to run the script when this flag is false
        return;
    }

    if (USE_WIDTH) {
        if (window.innerWidth == 0) {
            // We still shouldn't run the script becuase we are waiting for innerWidth to be non zero
            return;
        }        
    } else {
        if (window.innerHeight == 0) {
            // We still shouldn't run the script becuase we are waiting for innerHeight to be non zero
            return;
        }
    }

    shouldRunOnResize = false;
    normalizeAllMessageWidths();
    	
}

// This function prepears the html for the future run
// It wraps the current body with a new div with a new class
function prepear() {
    var bodys = document.getElementsByTagName("body");
    var body = bodys[0];
    var original = body.innerHTML;

    var newHtml = "<div id='111' class='mail-message'><div class='mail-message-content'>" + original + "</div></div>";

    body.innerHTML = newHtml;

    console.log("prepeared HTML");
}

// Subscribe the page events
function subscribeEvents() {
    window.onload = onPageLoaded;
    window.onhaschange = function () { console.log("onhaschange w=" + window.innerWidth + " h=" + window.innerHeight); }
    window.onpageshow = function () { console.log("onpageshow w=" + window.innerWidth + " h=" + window.innerHeight); }
    window.onresize = onPageResize;

    console.log("subscribed the window events");
}

// This function will try to normilize the html directly after the execution in order 
function tryNormalizeOnBegin() {
    console.log("tryNormalizeOnBegin()- window size: w=" + window.innerWidth + " h=" + window.innerHeight);

    // 1) We need to run the script only if the webview (window) already has a width. If the width is 0 we can't start the resizing because we can't
    // know what is the size of the internal divs and bodies
    // 2) We also check the shouldRunOnResize flag just in case the onReisze callback was called before this function (Probably not impossible)
	// 3) Also check if it is a show pictures mode. If we are not in pictures mode there is no need to run the script on the beginning. We can wait 
	// to the onload event
    if (window.innerWidth > 0 && shouldRunOnResize && SHOW_PICTURES) {        
        normalizeAllMessageWidths();
    }
}

//========== MAIN ===============
// The main section of the script

/*
External parameters (Initialized inside the HTML):
=================================================

    int WEBVIEW_WIDTH - The width of the webview 
	boolean SHOW_PICTURES - Is in show pictures mode
	
	mail - is a JS bridge to Android web view

Knowlage:
=========
    1) Running the script when the width of the browser (window.innerWidth) is 0 cause a lot of problems. We need to listen to the resize event in order to 
    understand when we can run the script

    2) Resizing images using javasctipt after a long delay cause them to disappear (sometimes)


The script flow:
================

	== Main ==
    1. Wrap the body of the html with main div
	2. Subscribe to page events
	3. Try to normilize the html on the begging
		3.1 IF the width of the window is more than 0 (Never run the normalization script when the browser width is 0) 
			AND the script wasn't started by resize callback
			AND in case we are in show pictures (No need to run the normalization script on start when the SHOW_PICTURES is false because it will be 
			called very soon from onload callback. This will help us avoid rendering issues and broken UI)
			3.1.1 execute normalization script
			
	== OnLoad ==
		1. IF the width of the window is more than 0 AND the window size was changed 
			1.1 IF NOT in SHOW_PICUTERS mode
				1.1.1 reset the shouldRunOnResize flag because we don't need to run the normalization script on the next resize (No pictures will 
				downloaded and that's why anything will changed == reduce the case of artifacts)
			1.2 execute normalization script
	
	== OnResize ==
		1. IF shouldRunOnResize
			1.1 Reset shouldRunOnResize
			1.2 execute normalization script
			
	== Normalization script ==
		1. Select the main div
		2. IF NOT should be normalized (div.scroll > WEBVIEW_WIDTH * 0.7)
		2.1 return
		3. 

*/

var SHOW_PICTURES = mail.isDisplayPictures();
var USE_WIDTH = mail.shouldUseWidthMode();

console.log("script.js started (first line): WEBVIEW_WIDTH=" + WEBVIEW_WIDTH + "; SHOW_PICTURES=" + SHOW_PICTURES + " ; USE_WIDTH=" + USE_WIDTH);

// Prepear the html
prepear();

// Subscribe the page events
subscribeEvents();

tryNormalizeOnBegin();
